/*
 * Copyright IBM Corp. 2003,2008
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.ibm.com/developerworks/library/os-cpl.html
 *
 * Authors:		Sven Schuetz <sven@de.ibm.com>
 * 
 * File:		vlib_sysfs.c
 * 
 * Description:
 * All calls that use the sysfs and call into libsysfs.
 * 
 */
 
 /**
 * @file vlib_sysfs.c
 * @brief All calls that use the sysfs and call into libsysfs.
 */

#include <sysfs/libsysfs.h>

#include "vlib.h"

/* internal helper functions */

/**
 * @brief add a  port to the adapters repos
 * @param *adapter the adapter to which the add the port to
 * @param *name the unique name of the remote port as in fc_remote_ports
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 * @par Locks:
 * 	vlib_data.mutex must be held
 */
static HBA_STATUS addPortByName(struct vlib_adapter *adapter, char* name)
{
	struct sysfs_class_device *class_dev;
	struct sysfs_attribute *attr;
	struct vlib_port port;
	
	class_dev = sysfs_open_class_device("fc_remote_ports", name);
	if (class_dev == NULL)
		return HBA_STATUS_ERROR;
	
	strcpy(port.name, name);
	sscanf(name, "rport-%d:%d-%d", &port.host, &port.channel, &port.target);
	
	attr = sysfs_get_classdev_attr(class_dev, "node_name");
	if (attr != NULL)
		port.wwnn = strtol(attr->value, NULL, 16);
	attr = sysfs_get_classdev_attr(class_dev, "port_name");
	if (attr != NULL)
		port.wwpn = strtol(attr->value, NULL, 16);
	attr = sysfs_get_classdev_attr(class_dev, "port_id");
	if (attr != NULL)
		port.did = strtol(attr->value, NULL, 16);
	
	addPortToRepos(adapter, &port);

	sysfs_close_class_device(class_dev);
	return HBA_STATUS_OK;
}

/**
 * @brief add an adapter to the adapters repos
 * @param *bus_dev the sysfs bus device as seen under
 * 		/sys/bus/ccw/drivers/zfcp/x.x.xxxx
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 * @par Locks:
 * 	vlib_data.mutex must be held
 */
static HBA_STATUS addAdapterByBusDev(struct sysfs_device *bus_dev)
{
	struct dlist *dir_list;
	char *fc_host_name;
	struct sysfs_class_device *class_dev;
	struct vlib_adapter a;
	struct sysfs_attribute *attr;
	char path[SYSFS_PATH_MAX];
	
	attr = sysfs_get_device_attr(bus_dev, "online");
	if (attr && strncmp(attr->value, "0", 1) == 0) {
		return HBA_STATUS_ERROR_INVALID_HANDLE;
	}
        dir_list = sysfs_open_directory_list(bus_dev->path);
        if (dir_list == NULL) {
        	return HBA_STATUS_OK;
        }
        dlist_for_each_data(dir_list, fc_host_name, char) {
        	if (strncmp(fc_host_name, "host", 4) == 0) {
        		break;
        	}
        }
        if (fc_host_name == NULL) {
        	/* adapter is offline, no fc_host entry */
        	sysfs_close_list(dir_list);
        	return HBA_STATUS_ERROR_UNAVAILABLE;
        }
	class_dev = sysfs_open_class_device("fc_host", fc_host_name);
        if (class_dev == NULL) {
        	/* adapter is offline, no fc_host entry */
        	sysfs_close_list(dir_list);
        	return HBA_STATUS_ERROR_UNAVAILABLE;
        }
	strcpy(a.ident.bus_dev_name, bus_dev->name);
	strcpy(a.ident.class_dev_name, class_dev->name);
	memcpy(&a.ident.devid, bus_dev->name, 8);
	attr = sysfs_get_classdev_attr(class_dev, "node_name");
	if (attr != NULL)
		a.ident.wwnn = strtol(attr->value, NULL, 16);
	attr = sysfs_get_classdev_attr(class_dev, "port_name");
	if (attr != NULL)
		a.ident.wwpn = strtol(attr->value, NULL, 16);
	
	addAdapterToRepos(&a);
	
	sysfs_close_class_device(class_dev);
	sysfs_close_list(dir_list);
	
	return HBA_STATUS_OK;
}

/**
 * @brief Retrieve adapter attributes.
 * @param **pPortattributes, HBA_ADAPTERATTRIBUTES to be filled
 * @param *adapter to work with
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 *
 * This function reads attributes from sysfs to fill in the required
 * information.
 */
static HBA_STATUS getPortAttributes(HBA_PORTATTRIBUTES **pPortattributes,
					struct sysfs_class_device *class_dev)
{
	struct sysfs_attribute *attr;
	
	/* Worldwide Port and Node Name */
	attr = sysfs_get_classdev_attr(class_dev, "node_name");
	if (attr != NULL)
		vlib_wwn_to_HBA_WWN(strtol(attr->value, NULL, 16),
					&(*pPortattributes)->NodeWWN);
	attr = sysfs_get_classdev_attr(class_dev, "port_name");
	if (attr != NULL)
		vlib_wwn_to_HBA_WWN(strtol(attr->value, NULL, 16),
					&(*pPortattributes)->PortWWN);
	
	/* PortFcId */
	attr = sysfs_get_classdev_attr(class_dev, "port_id");
	if (attr) {
		(*pPortattributes)->PortFcId = strtol(attr->value, NULL, 16);
		(*pPortattributes)->PortFcId =
			vlib_FCID_to_hbaFCID((*pPortattributes)->PortFcId);
	}

	/* Port Type */
	attr = sysfs_get_classdev_attr(class_dev, "port_type");
	if (attr)
		(*pPortattributes)->PortType =
					vlibCharToIntPortType(attr->value);
	
	/* Port State */
	attr = sysfs_get_classdev_attr(class_dev, "port_state");
	if (attr)
		(*pPortattributes)->PortState = 
					vlibCharToIntPortState(attr->value);
	
	/* Supported Classes */
	attr = sysfs_get_classdev_attr(class_dev, "supported_classes");
	if (attr)
		(*pPortattributes)->PortSupportedClassofService =
						vlibCharToIntCOS(attr->value);
	/* Supported FC4 types, we only support SCSI FCP which is
	 * represented by 0x0000 0100 in Word 1 */
	(*pPortattributes)->PortSupportedFc4Types.bits[2] = 0x1;
	
	/* 0 when port down, otherwise same as above */
	if((*pPortattributes)->PortState == HBA_PORTSTATE_ONLINE)
		(*pPortattributes)->PortActiveFc4Types.bits[2] = 0x1;
	
	/* Symbolic Name is empty */
	
	/* OSDeviceName is empty, we do not have a device */
	
	/* Supported port speeds */
	attr = sysfs_get_classdev_attr(class_dev, "supported_speeds");
	if (attr)
		(*pPortattributes)->PortSupportedSpeed =
					vlibCharToIntPortSpeed(attr->value);
	
	/* port speed */
	attr = sysfs_get_classdev_attr(class_dev, "speed");
	if (attr)
		(*pPortattributes)->PortSpeed =
					vlibCharToIntPortSpeed(attr->value);
	
	/* max frame size */
	attr = sysfs_get_classdev_attr(class_dev, "maxframe_size");
	if (attr)
		(*pPortattributes)->PortMaxFrameSize = atoi(attr->value);
	
	/* FabricName is empty */
	
	return HBA_STATUS_OK;
}

/* functions directly called from vlib.c or vlib_aux.c */


 /**
 * @brief Read and store all discovered ports of an adapter
 * @param *adapter pointer to the adapter in which we are interested
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 * @par Locks:
 * 	vlib_data.mutex must be held
 */
HBA_STATUS sysfs_createAndReadConfigPorts(struct vlib_adapter *adapter)
{
	HBA_STATUS status;
	struct sysfs_class_device *class_dev;
	struct sysfs_class_device *class_dev_leafs;
	struct dlist *dir_list;
	char path[256];
	char *leafname;
        
        if(adapter->ident.devid == 0)
        	return HBA_STATUS_ERROR;
        
	class_dev = sysfs_open_class_device("fc_host",
						adapter->ident.class_dev_name);	
	strcpy(path, class_dev->path);
	strcat(path, "/device");
	class_dev_leafs = sysfs_open_class_device_path(path);
	
	dir_list = sysfs_open_directory_list(class_dev_leafs->path);
	if (dir_list != NULL) {
		dlist_for_each_data(dir_list, leafname, char) {
			if (strncmp(leafname, "rport", 5) == 0) {
				addPortByName(adapter, leafname);
			}
		}
	}
	
	sysfs_close_class_device(class_dev);
	sysfs_close_class_device(class_dev_leafs);
	
	return HBA_STATUS_OK;
}

/**
 * @brief Read all adapters from /sys/bus/ccw/drivers/zfcp and add them
 * 	to the repository
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 * @par Locks:
 * 	vlib_data.mutex must be held
 */
HBA_STATUS sysfs_createAndReadConfigAdapter()
{
	HBA_STATUS status;
	int ret;
        struct dlist *dev_list;
        struct sysfs_bus *bus;
        struct sysfs_device *bus_dev;
        struct sysfs_driver *driver;
        
	bus = sysfs_open_bus("ccw");
        if (bus == NULL) {
        	return HBA_STATUS_OK;
        }
        driver = sysfs_get_bus_driver(bus, "zfcp");
        if (driver == NULL) {
        	sysfs_close_bus(bus);
        	return HBA_STATUS_OK;
        }

	dev_list = sysfs_get_driver_devices(driver);
        if (dev_list == NULL) {
        	sysfs_close_bus(bus);
        	return HBA_STATUS_OK;
        }
        
        dlist_for_each_data(dev_list, bus_dev, struct sysfs_device) {
        	addAdapterByBusDev(bus_dev);
        }
        
       	ret = revalidateAdapters();
	if (ret < 0) {
		status = HBA_STATUS_ERROR;
	} else {
		vlib_data.isValid = 1;
		status = HBA_STATUS_OK;
	}

	sysfs_close_bus(bus);

	return status;
}

/**
 * @brief Get unit configuration information for a port.
 * @param *port for which unit configuration is received
 * @return
 *	- -1 on error
 *	- 0 on success.
 * @par Locks:
 *	vlib_data.mutex must be held
 *
 * This function creates and reads private configuration events for all
 * configured units of an port. It enlarges the array where unit data is
 * stored and processes the generated configuration events.
 */
int sysfs_getUnitsFromPort(struct vlib_port *port)
{
	char path[256];
	char dev_path[256];
	struct sysfs_class_device *class_dev;
	struct sysfs_class_device *scsi_dev;
	struct sysfs_attribute *attr;
	struct dlist *dir_list;
	struct dlist *dlist_sg;
	char* dirname;
	char* sg;
	struct vlib_unit unit;
	
	class_dev = sysfs_open_class_device("fc_remote_ports", port->name);
	if(class_dev == NULL) {
		/* error, fc_remote_ports port n/a */
		return -1;
	}
	strcpy(path, class_dev->path);
	sysfs_close_class_device(class_dev);
	strcat(path, "/device");
	
	class_dev = sysfs_open_class_device_path(path);
	dir_list = sysfs_open_directory_list(class_dev->path);
	if (dir_list != NULL) {
		dlist_for_each_data(dir_list, dirname, char) {
			if (strncmp(dirname, "target", 6) == 0) {
				break;
			}
		}
	} else {
		/* no units, no error */
		return 0;
	}
	if(dirname == NULL) {
		/* no units, no error */
		return 0;
	}
	strcat(path, "/");
	strcat(path, dirname);
	sysfs_close_list(dir_list);
	sysfs_close_class_device(class_dev);
	
	class_dev = sysfs_open_class_device_path(path);
	dir_list = sysfs_open_directory_list(class_dev->path);
	if (dir_list == NULL) {
		/* no units */
		sysfs_close_list(dir_list);
		sysfs_close_class_device(class_dev);
		return 0;
	}
	dlist_for_each_data(dir_list, dirname, char) {
		strcpy(dev_path, path);
		strcat(dev_path, "/");
		strcat(dev_path, dirname);
		scsi_dev = sysfs_open_class_device_path(dev_path);
		sscanf(dirname, "%d:%d:%d:%d", &unit.host, &unit.channel,
						&unit.target, &unit.lun);
		attr = sysfs_get_classdev_attr(scsi_dev, "fcp_lun");
		if(attr)
			unit.fcLun = strtol(attr->value, NULL, 16);
		unit.sg_dev[0] = '\0';
		
		// test begin
		sysfs_close_class_device(scsi_dev);
		strcat(dev_path, "/scsi_generic");
		scsi_dev = sysfs_open_class_device_path(dev_path);
		if (!scsi_dev)
			break;
		dlist_sg = sysfs_open_directory_list(scsi_dev->path);
		if(dlist_sg) {
			dlist_for_each_data(dlist_sg, sg, char) {
				sscanf(sg, "%s", &unit.sg_dev);
			}
			sysfs_close_list(dlist_sg);
		}
		// test end
		/*dlist_sg = sysfs_open_directory_list(scsi_dev->path);
		if(dlist_sg) {
			dlist_for_each_data(dlist_sg, sg, char) {
				if(sscanf(sg, "scsi_generic:%s", &unit.sg_dev))
					break;
			}
		}*/
		addUnitToRepos(port, &unit);
		sysfs_close_class_device(scsi_dev);
	}
	
	sysfs_close_class_device(class_dev);
	sysfs_close_list(dir_list);
	return 0;
}

/**
 * @brief Retrieve adapter attributes.
 * @param **pPortattributes, HBA_ADAPTERATTRIBUTES to be filled
 * @param *adapter to work with
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 *
 * This function reads attributes from sysfs to fill in the required
 * information.
 */
HBA_STATUS sysfs_getDiscoveredPortAttributes(HBA_PORTATTRIBUTES **pAttrs,
						struct vlib_port *port)
{
	struct sysfs_class_device *class_dev;
	
	memset(*pAttrs, 0, sizeof(HBA_PORTATTRIBUTES));
	
	class_dev = sysfs_open_class_device("fc_remote_ports", port->name);
	
	if(! class_dev)
		return HBA_STATUS_ERROR_UNAVAILABLE;

	getPortAttributes(pAttrs, class_dev);
	
	/* not applicable to remote ports at the moment */
	memset(&(*pAttrs)->PortActiveFc4Types, 0, sizeof(HBA_FC4TYPES));
	memset(&(*pAttrs)->PortSupportedFc4Types, 0, sizeof(HBA_FC4TYPES));
	
	sysfs_close_class_device(class_dev);
	
	return HBA_STATUS_OK;
}

/**
 * @brief Retrieve adapter attributes.
 * @param **pPortattributes, HBA_ADAPTERATTRIBUTES to be filled
 * @param *adapter to work with
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 *
 * This function reads attributes from sysfs to fill in the required
 * information.
 */
HBA_STATUS sysfs_getAdapterPortAttributes(HBA_PORTATTRIBUTES **pAttrs,
						struct vlib_adapter *adapter)
{
	struct sysfs_class_device *class_dev;
	struct sysfs_class_device *class_dev_leafs;
	struct sysfs_attribute *attr;
	struct dlist *dir_list;
	char path[256];
	char *leafname;
	
        if(adapter->ident.devid == 0)
        	return HBA_STATUS_ERROR_UNAVAILABLE;
        	
	memset(*pAttrs, 0, sizeof(HBA_PORTATTRIBUTES));
	
	class_dev = sysfs_open_class_device("fc_host",
						adapter->ident.class_dev_name);
	
	if(! class_dev)
		return HBA_STATUS_ERROR_UNAVAILABLE;

	getPortAttributes(pAttrs, class_dev);
	
	strcpy(path, class_dev->path);
	strcat(path, "/device");
	class_dev_leafs = sysfs_open_class_device_path(path);
	if(class_dev_leafs == NULL) {
		return HBA_STATUS_ERROR_UNAVAILABLE;
	}
	dir_list = sysfs_open_directory_list(class_dev_leafs->path);
	if (dir_list != NULL) {
		dlist_for_each_data(dir_list, leafname, char) {
			if (strncmp(leafname, "rport", 5) == 0) {
				(*pAttrs)->NumberofDiscoveredPorts++;
				addPortByName(adapter, leafname);
			}
		}
	}
	
	sysfs_close_list(dir_list);
	sysfs_close_class_device(class_dev_leafs);
	sysfs_close_class_device(class_dev);
	
	return HBA_STATUS_OK;
}

/**
 * @brief Retrieve adapter attributes.
 * @param **pAdapterattributes, HBA_ADAPTERATTRIBUTES to be filled
 * @param *adapter to work with
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 *
 * This function reads attributes from sysfs to fill in the required
 * information.
 */
HBA_STATUS sysfs_getAdapterAttributes(HBA_ADAPTERATTRIBUTES **pAttrs,
						struct vlib_adapter *adapter)
{
	struct sysfs_device *bus_dev;
	struct sysfs_driver *driver;
	struct sysfs_bus *bus;
	struct sysfs_class_device *class_dev;
	struct sysfs_attribute *attr;
	int a;
	
	
	bus = sysfs_open_bus("ccw");
	if (bus == NULL) {
		return HBA_STATUS_ERROR_INVALID_HANDLE;
	}
	memset(*pAttrs, 0, sizeof(HBA_ADAPTERATTRIBUTES));
	
	bus_dev = sysfs_get_bus_device(bus, adapter->ident.bus_dev_name);
	if (bus_dev == NULL) {
		return HBA_STATUS_ERROR_INVALID_HANDLE;
	}
	driver = sysfs_get_bus_driver(bus, "zfcp");
	if (driver == NULL) {
		return HBA_STATUS_ERROR_INVALID_HANDLE;
	}
	class_dev = sysfs_open_class_device("fc_host",
						adapter->ident.class_dev_name);
	if (class_dev == NULL) {
		return HBA_STATUS_ERROR_INVALID_HANDLE;
	}

        /* Manufacturer */
        strcpy((*pAttrs)->Manufacturer, "IBM");
        
        /* Serial Number */
        attr = sysfs_get_classdev_attr(class_dev, "serial_number");
        if (attr)
        	strcpy((*pAttrs)->SerialNumber, attr->value);
        
        /* Model */
        attr = sysfs_get_device_attr(bus_dev, "card_version");
        if (attr) {
	        a = strtol(attr->value, NULL, 16);
	        strcpy((*pAttrs)->ModelDescription,
	        		"zSeries/System z Fibre Channel Adapter ");
	        switch (a) {
	        case 1:
	        	strcpy((*pAttrs)->Model, "Ficon Adapater");
	        	strcat((*pAttrs)->ModelDescription, "Hydra 1.5");
	        	break;
	        case 2:
	        	strcpy((*pAttrs)->Model, "Ficon Express Adapater");
	        	strcat((*pAttrs)->ModelDescription, "Hydra 1.75");
	        	break;
	        case 3:
	        	strcpy((*pAttrs)->Model, "Ficon Express2 Adapater");
	        	strcat((*pAttrs)->ModelDescription,
	        				"Ficon-3 with 2 Gbit/s");
	        	break;
	        case 4:
	        	strcpy((*pAttrs)->Model, "Ficon Express2.5 Adapater");
	        	strcat((*pAttrs)->ModelDescription,
	        				"Ficon-3 with 4 Gbit/s");
	        	break;
	        }
        }
        /* World Wide Node Name */
	vlib_wwn_to_HBA_WWN(adapter->ident.wwnn, &(*pAttrs)->NodeWWN);
	
	/* NodeSymbolicName not set */
	
	/* Hardware Version */
	attr = sysfs_get_device_attr(bus_dev, "hardware_version");
	if (attr)
		strcpy((*pAttrs)->HardwareVersion, attr->value);
	
	/* Driver Version */
	attr = sysfs_get_driver_attr(driver, "version");
	if (attr)
		strcpy((*pAttrs)->DriverVersion, attr->value);
	
	/* OptionROMVersion not set */
	
	/* VendorSpecificID will be the devno a.b.cccc
	 * a will be stored in the first byte
	 * b will be stored in the second byte
	 * cccc will be stored in the last two bytes */
	sscanf(adapter->ident.bus_dev_name, "%hhd.%hhd.%hx",
			(char*)&(*pAttrs)->VendorSpecificID,
			(char*)(&(*pAttrs)->VendorSpecificID) + 1,
			(char*)(&(*pAttrs)->VendorSpecificID) + 2);

	/* Firmware (here LIC) Version */
	attr = sysfs_get_device_attr(bus_dev, "lic_version");
	if (attr)
		strcpy((*pAttrs)->FirmwareVersion, attr->value);
	
	/* Number of adapters ... always one */
	(*pAttrs)->NumberOfPorts = 1; /* always 1 */
	
	/* Driver name */
	strcpy((*pAttrs)->DriverName, "zfcp");
	
        sysfs_close_class_device(class_dev);
        sysfs_close_bus(bus);
        return HBA_STATUS_OK;
}

/**
 * @brief Retrieve adapter port statistics
 * @param **pPortstatistics, HBA_PORTSTATISTICS to be filled
 * @param *adapter to work with
 * @return
 *	- HBA_STATUS_ERROR if any other internal error occurs
 * 	- HBA_STATUS_OK on success.
 *
 * This function reads attributes from sysfs to fill in the required
 * information.
 */
HBA_STATUS sysfs_getPortStatistics(HBA_PORTSTATISTICS **pS,
						struct vlib_adapter *adapter)
{

	struct sysfs_class_device *class_dev;
	struct sysfs_class_device *stats;
	struct sysfs_attribute *attr;
	char path[256];
	
	memset(*pS, 0, sizeof(HBA_PORTSTATISTICS));

	class_dev = sysfs_open_class_device("fc_host",
						adapter->ident.class_dev_name);

	strcpy(path, class_dev->path);
	strcat(path, "/statistics");

	stats = sysfs_open_class_device_path(path);
	
	if(class_dev == NULL || stats == NULL) {
		return HBA_STATUS_ERROR_UNAVAILABLE;
	}

	attr = sysfs_get_classdev_attr(stats, "seconds_since_last_reset");
	if(attr)
		(*pS)->SecondsSinceLastReset = strtol(attr->value,
								NULL, 16);
	attr = sysfs_get_classdev_attr(stats, "tx_frames");
	if(attr)
		(*pS)->TxFrames = strtol(attr->value, NULL, 16);

	attr = sysfs_get_classdev_attr(stats, "tx_words");
	if(attr)
		(*pS)->TxWords = strtol(attr->value, NULL, 16);
		
	attr = sysfs_get_classdev_attr(stats, "rx_frames");
	if(attr)
		(*pS)->RxFrames = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "rx_words");
	if(attr)
		(*pS)->RxWords = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "lip_count");
	if(attr)
		(*pS)->LIPCount = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "nos_count");
	if(attr)
		(*pS)->NOSCount = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "error_frames");
	if(attr)
		(*pS)->ErrorFrames = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "dumped_frames");
	if(attr)
		(*pS)->DumpedFrames = strtol(attr->value, NULL, 16);
						
	attr = sysfs_get_classdev_attr(stats, "link_failure_count");
	if(attr)
		(*pS)->LinkFailureCount = strtol(attr->value, NULL, 16);
		
	attr = sysfs_get_classdev_attr(stats, "loss_of_sync_count");
	if(attr)
		(*pS)->LossOfSyncCount = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "loss_of_signal_count");
	if(attr)
		(*pS)->LossOfSignalCount = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "prim_seq_protocol_err_count");
	if(attr)
		(*pS)->PrimitiveSeqProtocolErrCount = strtol(attr->value, NULL,
									16);
	
	attr = sysfs_get_classdev_attr(stats, "invalid_tx_word_count");
	if(attr)
		(*pS)->InvalidTxWordCount = strtol(attr->value, NULL, 16);
	
	attr = sysfs_get_classdev_attr(stats, "invalid_crc_count");
	if(attr)
		(*pS)->InvalidCRCCount = strtol(attr->value, NULL, 16);

        sysfs_close_class_device(stats);					
        sysfs_close_class_device(class_dev);
        
        return HBA_STATUS_OK;
}
